@@ -6,18 +6,18 @@ use crate::tokio::{
     time::timeout,
 };
 use crate::{
-    error::Error,
+    error::*,
     parser,
     relay::{CtrlSignal, ProcessorRelay, RelayMessage, WorkerRelay},
     router::SenderPair,
     Cancel, NodeMessage, ShutdownType,
 };
-use crate::{NodeError, Reason};
 use core::time::Duration;
 use ockam_core::compat::{boxed::Box, string::String, sync::Arc, vec::Vec};
 use ockam_core::{
-    AccessControl, Address, AddressSet, AllowAll, AsyncTryClone, LocalMessage, Message, Processor,
-    Result, Route, TransportMessage, TransportType, Worker,
+    errcode::{Kind, Origin},
+    AccessControl, Address, AddressSet, AllowAll, AsyncTryClone, Error, LocalMessage, Message,
+    Processor, Result, Route, TransportMessage, TransportType, Worker,
 };
 
 /// A default timeout in seconds
@@ -61,11 +61,9 @@ impl Context {
     /// Wait for the next message from the mailbox
     pub(crate) async fn mailbox_next(&mut self) -> Result<Option<RelayMessage>> {
         loop {
-            let relay_msg = if let Some(msg) = self.mailbox.recv().await.map(|msg| {
+            let relay_msg = if let Some(msg) = self.mailbox.recv().await {
                 trace!("{}: received new message!", self.address());
                 msg
-            }) {
-                msg
             } else {
                 return Ok(None);
             };
@@ -150,13 +148,11 @@ impl Context {
         self.sender
             .send(msg)
             .await
-            .map_err(|_| Error::FailedStartWorker)?;
-
-        Ok(rx
-            .recv()
+            .map_err(|e| Error::new(Origin::Node, Kind::Invalid, e))?;
+        rx.recv()
             .await
-            .ok_or(Error::InternalIOFailure)?
-            .map(|_| ctx)?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(ctx)
     }
 
     /// Start a new worker instance at the given address set
@@ -246,14 +242,13 @@ impl Context {
         self.sender
             .send(msg)
             .await
-            .map_err(|_| Error::FailedStartWorker)?;
+            .map_err(|e| Error::new(Origin::Node, Kind::Invalid, e))?;
 
         // Wait for the actual return code
-        Ok(rx
-            .recv()
+        rx.recv()
             .await
-            .ok_or(Error::InternalIOFailure)?
-            .map(|_| ())?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(())
     }
 
     /// Start a new processor instance at the given address set
@@ -287,14 +282,13 @@ impl Context {
         self.sender
             .send(msg)
             .await
-            .map_err(|_| Error::FailedStartProcessor)?;
+            .map_err(|e| Error::new(Origin::Node, Kind::Invalid, e))?;
 
         // Wait for the actual return code
-        Ok(rx
-            .recv()
+        rx.recv()
             .await
-            .ok_or(Error::InternalIOFailure)?
-            .map(|_| ())?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(())
     }
 
     /// Shut down a local worker by its primary address
@@ -315,14 +309,16 @@ impl Context {
             AddressType::Worker => NodeMessage::stop_worker(addr),
             AddressType::Processor => NodeMessage::stop_processor(addr),
         };
-        self.sender.send(req).await.map_err(Error::from)?;
+        self.sender
+            .send(req)
+            .await
+            .map_err(NodeError::from_send_err)?;
 
         // Then check that address was properly shut down
-        Ok(rx
-            .recv()
+        rx.recv()
             .await
-            .ok_or(Error::InternalIOFailure)?
-            .map(|_| ())?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(())
     }
 
     /// Signal to the local runtime to shut down immediately
@@ -337,7 +333,7 @@ impl Context {
 
         match tx.send(msg).await {
             Ok(()) => Ok(()),
-            Err(_e) => Err(Error::FailedStopNode.into()),
+            Err(e) => Err(Error::new(Origin::Node, Kind::Invalid, e)),
         }
     }
 
@@ -357,14 +353,16 @@ impl Context {
     /// or the desired timeout has been reached.
     pub async fn stop_timeout(&mut self, seconds: u8) -> Result<()> {
         let (req, mut rx) = NodeMessage::stop_node(ShutdownType::Graceful(seconds));
-        self.sender.send(req).await.map_err(Error::from)?;
+        self.sender
+            .send(req)
+            .await
+            .map_err(NodeError::from_send_err)?;
 
         // Wait until we get the all-clear
-        Ok(rx
-            .recv()
+        rx.recv()
             .await
-            .ok_or(Error::InternalIOFailure)?
-            .map(|_| ())?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(())
     }
 
     /// Send a message to another address associated with this worker
@@ -381,7 +379,7 @@ impl Context {
         if self.address.contains(&addr) {
             self.send_from_address(addr, msg, from.into()).await
         } else {
-            Err(NodeError::Rejected(Reason::InvalidAddress).into())
+            Err(NodeError::NodeState(NodeReason::Unknown).internal())
         }
     }
 
@@ -460,7 +458,7 @@ impl Context {
         M: Message + Send + 'static,
     {
         if !self.address.as_ref().contains(&sending_address) {
-            return Err(Error::SenderAddressDoesNotExist.into());
+            return Err(Error::new_without_cause(Origin::Node, Kind::Invalid));
         }
 
         let (reply_tx, mut reply_rx) = channel(1);
@@ -468,11 +466,14 @@ impl Context {
         let req = NodeMessage::SenderReq(next.clone(), reply_tx);
 
         // First resolve the next hop in the route
-        self.sender.send(req).await.map_err(Error::from)?;
+        self.sender
+            .send(req)
+            .await
+            .map_err(NodeError::from_send_err)?;
         let (addr, sender, needs_wrapping) = reply_rx
             .recv()
             .await
-            .ok_or(Error::InternalIOFailure)??
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??
             .take_sender()?;
 
         // Pack the payload into a TransportMessage
@@ -489,8 +490,7 @@ impl Context {
         };
 
         // Send the packed user message with associated route
-        sender.send(msg).await.map_err(Error::from)?;
-
+        sender.send(msg).await.map_err(NodeError::from_send_err)?;
         Ok(())
     }
 
@@ -513,11 +513,14 @@ impl Context {
         let req = NodeMessage::SenderReq(next.clone(), reply_tx);
 
         // First resolve the next hop in the route
-        self.sender.send(req).await.map_err(Error::from)?;
+        self.sender
+            .send(req)
+            .await
+            .map_err(NodeError::from_send_err)?;
         let (addr, sender, needs_wrapping) = reply_rx
             .recv()
             .await
-            .ok_or(Error::InternalIOFailure)??
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??
             .take_sender()?;
 
         // Pack the transport message into a relay message
@@ -528,7 +531,7 @@ impl Context {
         } else {
             RelayMessage::direct(addr, local_msg, onward)
         };
-        sender.send(msg).await.map_err(Error::from)?;
+        sender.send(msg).await.map_err(NodeError::from_send_err)?;
 
         Ok(())
     }
@@ -569,7 +572,7 @@ impl Context {
             self.next_from_mailbox().await
         })
         .await
-        .map_err(Error::from)??;
+        .map_err(|e| NodeError::Data.with_elapsed(e))??;
         Ok(Cancel::new(msg, data, addr, self))
     }
 
@@ -599,7 +602,7 @@ impl Context {
             }
         })
         .await
-        .map_err(Error::from)??;
+        .map_err(|e| NodeError::Data.with_elapsed(e))??;
 
         Ok(Cancel::new(m, data, addr, self))
     }
@@ -621,21 +624,30 @@ impl Context {
     /// initialisation when the node is stopped.
     pub async fn set_cluster<S: Into<String>>(&self, label: S) -> Result<()> {
         let (msg, mut rx) = NodeMessage::set_cluster(self.address(), label.into());
-        self.sender.send(msg).await.map_err(Error::from)?;
-        Ok(rx.recv().await.ok_or(Error::InternalIOFailure)??.is_ok()?)
+        self.sender
+            .send(msg)
+            .await
+            .map_err(NodeError::from_send_err)?;
+        rx.recv()
+            .await
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??
+            .is_ok()
     }
 
     /// Return a list of all available worker addresses on a node
     pub async fn list_workers(&self) -> Result<Vec<Address>> {
         let (msg, mut reply_rx) = NodeMessage::list_workers();
 
-        self.sender.send(msg).await.map_err(Error::from)?;
+        self.sender
+            .send(msg)
+            .await
+            .map_err(NodeError::from_send_err)?;
 
-        Ok(reply_rx
+        reply_rx
             .recv()
             .await
-            .ok_or(Error::InternalIOFailure)??
-            .take_workers()?)
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??
+            .take_workers()
     }
 
     /// Register a router for a specific address type
@@ -648,7 +660,7 @@ impl Context {
         self.sender
             .send(NodeMessage::StopAck(self.address()))
             .await
-            .map_err(Error::from)?;
+            .map_err(NodeError::from_send_err)?;
         Ok(())
     }
 
@@ -657,9 +669,12 @@ impl Context {
         self.sender
             .send(NodeMessage::Router(type_, addr, tx))
             .await
-            .map_err(|_| Error::InternalIOFailure)?;
+            .map_err(NodeError::from_send_err)?;
 
-        Ok(rx.recv().await.ok_or(Error::InternalIOFailure)??.is_ok()?)
+        rx.recv()
+            .await
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
+        Ok(())
     }
 
     /// A convenience function to get a data 3-tuple from the mailbox
@@ -677,7 +692,10 @@ impl Context {
     /// has woken it.
     async fn next_from_mailbox<M: Message>(&mut self) -> Result<(M, LocalMessage, Address)> {
         loop {
-            let msg = self.mailbox_next().await?.ok_or(Error::FailedLoadData)?;
+            let msg = self
+                .mailbox_next()
+                .await?
+                .ok_or_else(|| NodeError::Data.not_found())?;
             let (addr, data) = msg.local_msg();
 
             // FIXME: make message parsing idempotent to avoid cloning
@@ -701,7 +719,7 @@ impl Context {
         self.sender
             .send(NodeMessage::set_ready(self.address()))
             .await
-            .map_err(|_| Error::InternalIOFailure)?;
+            .map_err(NodeError::from_send_err)?;
         Ok(())
     }
 
@@ -711,11 +729,14 @@ impl Context {
         self.sender
             .send(msg)
             .await
-            .map_err(|_| Error::InternalIOFailure)?;
+            .map_err(NodeError::from_send_err)?;
 
         // This call blocks until the address has become ready or is
-        // dropped
-        reply.recv().await.ok_or(Error::InternalIOFailure)??;
+        // dropped by the router
+        reply
+            .recv()
+            .await
+            .ok_or_else(|| NodeError::NodeState(NodeReason::Unknown).internal())??;
         Ok(())
     }
 }
